/*
 *  glst by Davide Libenzi (SMTP GreyListing module)
 *  Copyright (C) 2004,..,2008  Davide Libenzi
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Davide Libenzi <davidel@xmailserver.org>
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <stdarg.h>
#include <gdbm.h>


#if defined(WIN32)

#include <winsock2.h>
#include <ws2tcpip.h>
#include <mswsock.h>

#define SLASH_CHAR '\\'
#define SLASH_STR "\\"
#define strncasecmp _strnicmp
#define strcasecmp _stricmp
#define snprintf _snprintf
#define vsnprintf _vsnprintf

#else /* defined(WIN32) */

#include <netdb.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <netinet/in.h>

#define SLASH_CHAR '/'
#define SLASH_STR "/"

#endif /* defined(WIN32) */

#define CSTRSIZE(s) (sizeof(s) - 1)

#define STRSPRINTF(res, last, fmt)					\
	do {								\
		int csize = 512, psize;					\
		va_list args;						\
		for (;;) {						\
			res = (char *) malloc(csize);			\
			if (res == NULL) break;				\
			va_start(args, last);				\
			if ((psize = vsnprintf(res, csize - 1, fmt, args)) >= 0 && \
			    psize < csize) {				\
				va_end(args);				\
				break;					\
			}						\
			va_end(args);					\
			if (psize > 0)					\
				csize = (4 * psize) / 3 + 2;		\
			else						\
				csize *= 2;				\
			free(res);					\
		}							\
	} while (0)

#define GLST_IN4(a) ((struct sockaddr_in *) (a))
#define GLST_IN6(a) ((struct sockaddr_in6 *) (a))

#define MAIL_DATA_TAG "<<MAIL-DATA>>"
#define MAIL_FROM "MAIL FROM:"
#define RCPT_TO "RCPT TO:"

#define GLST_OP_GRAYLIST 1
#define GLST_OP_CLEANUP 2

#define GLST_INET46 -1
#define GLST_INET64 -2

#define GLST_GENERROR 1
#define GLST_REJERROR 2
#define GLST_DEF_TIMEO 1200
#define GLST_DEF_EXPTIMEO (3600 * 24 * 36)
#define GLST_DEF_LAMETIMEO (3600 * 8)
#define GLST_DEF_REJMSG "451 4.7.1 Please try again later"
#define GLST_CONFIG_FILE "glst.conf"
#define GLST_DB_FILE "glst.dbm"
#define GLST_LAMEDB_FILE "glst-lame.dbm"



typedef struct s_glst_glrecord {
	time_t rtime;
	time_t mtime;
	int mcount;
	int acount;
} glst_glrecord_t;

typedef struct s_glst_lmrecord {
	time_t rtime;
	time_t mtime;
	int count;
} glst_lmrecord_t;

typedef struct s_glst_sockaddr {
	struct sockaddr a;
	unsigned char adata[64];
} glst_sockaddr_t;

typedef struct s_glst_subnet {
	struct s_glst_subnet *next;
	glst_sockaddr_t net;
	glst_sockaddr_t mask;
} glst_subnet_t;

typedef struct s_glst_masknet {
	struct s_glst_masknet *next;
	glst_sockaddr_t net;
	glst_sockaddr_t mask;
	glst_sockaddr_t nmask;
} glst_masknet_t;

typedef struct s_glst_ptrchain {
	struct s_glst_ptrchain *next;
	char *ptr;
} glst_ptrchain_t;

typedef struct s_glst_cfg {
	int generror;
	int rejerror;
	int timeo;
	int exptimeo;
	int lametimeo;
	char *root;
	char *rejmsg;
	glst_subnet_t *xnets;
	glst_masknet_t *mnets;
} glst_cfg_t;

typedef struct s_glst_ctx {
	int op;
	char *mfile, *rejfile;
	char *raddr, *mraddr, *laddr;
	char *sender;
	glst_ptrchain_t *rcpts;
	glst_ptrchain_t *crcpt;
} glst_ctx_t;

typedef struct s_glst_keychain {
	struct s_glst_keychain *next;
	char *key;
	char *data;
} glst_keychain_t;


static char *glst_sprintf(char const *fmt, ...);
static char *glst_memdup(char const *data, int size, int extra);
static char *glst_strchr(char const *str, int size, int ch);
static char *glst_strrchr(char const *str, int size, int ch);
static int glst_get_address(char const *data, int dsize, char const **addr, int *asize);
static char *glst_deVERP_address(char const *addr, int size);
static char const *glst_nextline(char const *pos, char const **next);
static int glst_gethostbyname(char const *name, int family, int isip,
			      glst_sockaddr_t *addr);
static void *glst_addr_data(glst_sockaddr_t const *addr, int *size);
static int glst_addr_size(glst_sockaddr_t const *addr);
static int glst_addr_6compat4(glst_sockaddr_t const *addr);
static int glst_addr_match(glst_sockaddr_t const *addr, glst_sockaddr_t const *net,
			   glst_sockaddr_t const *mask);
static char *glst_inet_ntoa(glst_sockaddr_t const *addr, char *str, int size);
static int glst_is_ip(char const *str);
static int glst_parse_ipaddr(char const *str, char *addr, int asize, int *port);
static int glst_parse_xmail_addrline(char const *line, int size, glst_cfg_t *cfg,
				     glst_ctx_t *ctx);
static int glst_complete_ctx(char const *hblk, glst_cfg_t *cfg, glst_ctx_t *ctx);
static char *glst_ctx_first_key(glst_ctx_t *ctx);
static char *glst_ctx_next_key(glst_ctx_t *ctx);
static int glst_is_complete_ctx(glst_ctx_t const *ctx);
static int glst_add_rcpt(glst_ctx_t *ctx, char const *rcpt, int size);
static int glst_add_exnet(glst_cfg_t *cfg, char const *xnet, int size);
static int glst_on_xnet(glst_cfg_t *cfg, glst_ctx_t *ctx);
static int glst_init_ctx(glst_ctx_t *ctx);
static char *glst_load_hdrs(FILE *fin);
static int glst_add_masknet(glst_cfg_t *cfg, char const *mnet, int size);
static char *glst_get_masked_raddr(glst_cfg_t *cfg, char const *raddr);
static int glst_assign_raddr(glst_cfg_t *cfg, glst_ctx_t *ctx, char const *raddr);
static int glst_set_raddr(glst_cfg_t *cfg, glst_ctx_t *ctx, char const *raddr);
static int glst_init(int ac, char **av);
static int glst_init_config(char const *prg, glst_cfg_t *cfg);
static int glst_load_config(char const *prg, char const *fcfg, glst_cfg_t *cfg);
static int glst_get_dbvar(char const *data, int dsize, char const *name,
			  char *vbuf, int vsize);
static int glst_init_glrecord(glst_glrecord_t *rec);
static void glst_free_glrecord(glst_glrecord_t *rec);
static int glst_parse_glrecord(datum ddata, glst_glrecord_t *rec);
static int glst_insert_glrecord(GDBM_FILE gdbf, datum dkey, glst_glrecord_t const *rec);
static int glst_init_lmrecord(glst_lmrecord_t *rec);
static void glst_free_lmrecord(glst_lmrecord_t *rec);
static int glst_parse_lmrecord(datum ddata, glst_lmrecord_t *rec);
static int glst_insert_lmrecord(GDBM_FILE gdbf, datum dkey, glst_lmrecord_t const *rec);
static int glst_handle_recipient(GDBM_FILE gdbf, char *ckey, glst_ctx_t *ctx,
				 glst_cfg_t *cfg);
static int glst_handle_request(glst_ctx_t *ctx, glst_cfg_t *cfg);
static int glst_add_keychain(glst_keychain_t **keys, char const *key, char const *data);
static void glst_free_keychain(glst_keychain_t *kv);
static void glst_free_keychain_list(glst_keychain_t *keys);
static int glst_update_lamedb(glst_cfg_t *cfg, glst_keychain_t *keys);
static int glst_handle_cleanup(glst_cfg_t *cfg);
static void glst_usage(char const *prg);



static char *glst_sprintf(char const *fmt, ...) {
	char *str;

	STRSPRINTF(str, fmt, fmt);
	return str;
}

static char *glst_memdup(char const *data, int size, int extra) {
	char *ddup;

	if (size < 0)
		size = strlen(data);
	if ((ddup = (char *) malloc(size + extra + 1)) != NULL) {
		memcpy(ddup, data, size);
		ddup[size] = 0;
	}
	return ddup;
}

static char *glst_strchr(char const *str, int size, int ch) {

	for (; size > 0; str++, size--)
		if ((int) *str == ch)
			return (char *) str;
	return NULL;
}

static char *glst_strrchr(char const *str, int size, int ch) {

	for (str += size - 1; size > 0; str--, size--)
		if ((int) *str == ch)
			return (char *) str;
	return NULL;
}

static int glst_get_address(char const *data, int dsize, char const **addr, int *asize) {
	int i, s, e;
	char const *noads = " \t\r\n,:;'\"<>@/\\", *tmp;

	if ((tmp = glst_strrchr(data, dsize, '@')) == NULL)
		return -1;
	i = tmp - data;
	for (s = i - 1; s > 0 && !strchr(noads, data[s - 1]); s--);
	for (e = i + 1; e < (dsize - 1) && !strchr(noads, data[e + 1]); e++);
	if (e - i < 2)
		return -1;
	*addr = data + s;
	*asize = e - s + 1;
	return 0;
}

static char *glst_deVERP_address(char const *addr, int size) {
	int usize, dsize;
	char const *atp, *plusp;
	char *raddr;

	if ((atp = glst_strrchr(addr, size, '@')) == NULL ||
	    (plusp = glst_strchr(addr, size, '+')) == NULL ||
	    plusp > atp)
		return glst_memdup(addr, size, 0);
	usize = (int) (plusp - addr);
	dsize = (int) ((addr + size) - (atp + 1));
	if ((raddr = (char *) malloc(usize + dsize + 2)) == NULL)
		return NULL;
	memcpy(raddr, addr, usize);
	memcpy(raddr + usize, atp, dsize + 1);
	raddr[usize + dsize + 1] = '\0';

	return raddr;
}

static char const *glst_nextline(char const *pos, char const **next) {
	char const *eol = strchr(pos, '\n');

	if (eol == NULL)
		return NULL;
	if (*next != NULL)
		pos = *next + 1;
	*next = eol;

	return pos;
}

static int glst_gethostbyname(char const *name, int family, int isip,
			      glst_sockaddr_t *addr) {
	struct addrinfo *cres, *res, *res4 = NULL, *res6 = NULL;
	struct addrinfo ahints;

	memset(&ahints, 0, sizeof(ahints));
	ahints.ai_family = AF_UNSPEC;
	ahints.ai_flags = isip ? AI_NUMERICHOST: 0;
	if (getaddrinfo(name, NULL, &ahints, &res))
		return -1;
	for (cres = res; cres != NULL; cres = cres->ai_next) {
		if (cres->ai_addr->sa_family == AF_INET)
			res4 = cres;
		else if (cres->ai_addr->sa_family == AF_INET6)
			res6 = cres;
	}
	switch (family) {
	case AF_INET:
		cres = res4;
		break;
	case AF_INET6:
		cres = res6;
		break;
	case GLST_INET46:
		if ((cres = res4) == NULL)
			cres = res6;
		break;
	case GLST_INET64:
		if ((cres = res6) == NULL)
			cres = res4;
		break;
	default:
		cres = NULL;
	}
	if (cres != NULL) {
		if (cres->ai_addrlen > sizeof(*addr)) {
			freeaddrinfo(res);
			return -1;
		}
		memset(addr, 0, sizeof(*addr));
		memcpy(addr, cres->ai_addr, cres->ai_addrlen);
	}
	freeaddrinfo(res);

	return cres != NULL ? 0: -1;
}

static void *glst_addr_data(glst_sockaddr_t const *addr, int *size) {
	switch (addr->a.sa_family) {
	case AF_INET:
		*size = sizeof(GLST_IN4(addr)->sin_addr);
		return (void *) &GLST_IN4(addr)->sin_addr;

	case AF_INET6:
		*size = sizeof(GLST_IN6(addr)->sin6_addr);
		return (void *) &GLST_IN6(addr)->sin6_addr;
	}

	return NULL;
}

static int glst_addr_size(glst_sockaddr_t const *addr) {
	switch (addr->a.sa_family) {
	case AF_INET:
		return sizeof(struct sockaddr_in);

	case AF_INET6:
		return sizeof(struct sockaddr_in6);
	}

	return 0;
}

static int glst_addr_6compat4(glst_sockaddr_t const *addr) {
	int i, size;
	unsigned char const *adata;

	if (addr->a.sa_family != AF_INET6 ||
	    (adata = (unsigned char const *) glst_addr_data(addr, &size)) == NULL)
		return 0;
	for (i = 0; i < 10; i++)
		if (adata[i])
			return 0;

	return (adata[i] == 0xff && adata[i + 1] == 0xff) ||
		(adata[i] == 0 && adata[i + 1] == 0);
}

static int glst_addr_match(glst_sockaddr_t const *addr, glst_sockaddr_t const *net,
			   glst_sockaddr_t const *mask) {
	int asize, nsize, msize;
	unsigned char const *adata, *ndata, *mdata;

	if ((adata = (unsigned char const *) glst_addr_data(addr, &asize)) == NULL ||
	    (ndata = (unsigned char const *) glst_addr_data(net, &nsize)) == NULL ||
	    (mdata = (unsigned char const *) glst_addr_data(mask, &msize)) == NULL)
		return 0;
	if (addr->a.sa_family != net->a.sa_family) {
		if (addr->a.sa_family == AF_INET) {
			if (net->a.sa_family != AF_INET6 ||
			    !glst_addr_6compat4(net))
				return 0;
			ndata += nsize - asize;
			mdata += msize - asize;
			nsize = asize;
			msize = asize;
		} else if (addr->a.sa_family == AF_INET6) {
			if (net->a.sa_family != AF_INET ||
			    !glst_addr_6compat4(addr))
				return 0;
			adata += asize - nsize;
			asize = nsize;
		} else
			return 0;
	}
	for (; asize > 0; asize--, adata++, ndata++, mdata++)
		if ((*adata & *mdata) != (*ndata & *mdata))
			return 0;

	return 1;
}

static char *glst_inet_ntoa(glst_sockaddr_t const *addr, char *str, int size) {
	int asize = 0;

	*str = 0;
	getnameinfo(&addr->a, glst_addr_size(addr), str, size, NULL, 0,
		    NI_NUMERICHOST);

	return str;
}

static int glst_is_ip(char const *str) {
	glst_sockaddr_t addr;

	return glst_gethostbyname(str, GLST_INET46, 1, &addr) == 0 ? 1: 0;
}

static int glst_parse_ipaddr(char const *str, char *addr, int asize, int *port) {
	int alen;
	char const *aend, *tmp;

	if (*str == '[') {
		str++;
		if ((aend = strchr(str, ']')) == NULL)
			return -1;
		if ((tmp = strchr(aend + 1, '|')) == NULL)
			tmp = strchr(aend + 1, ':');
		if (tmp != NULL)
			*port = atoi(tmp + 1);
	} else {
		if ((tmp = strchr(str, '|')) == NULL)
			tmp = strchr(str, ':');
		if ((aend = tmp) != NULL)
			*port = atoi(tmp + 1);
	}
	if (aend != NULL) {
		alen = aend - str;
		if (alen > asize - 1)
			alen = asize - 1;
		strncpy(addr, str, alen);
		addr[alen] = 0;
	} else {
		strncpy(addr, str, asize - 1);
		addr[asize - 1] = 0;
	}

	return glst_is_ip(addr) ? 0: -1;
}

static int glst_parse_xmail_addrline(char const *line, int size, glst_cfg_t *cfg,
				     glst_ctx_t *ctx) {
	int sport, cport;
	char const *tokstr = ";";
	char *ldup, *cldom, *claddr, *svdom, *svaddr;
	char saddr[256], caddr[256];

	if ((ldup = glst_memdup(line, size, 0)) == NULL)
		return -1;
	if ((cldom = strtok(ldup, tokstr)) == NULL ||
	    (claddr = strtok(NULL, tokstr)) == NULL) {
		free(ldup);
		return -1;
	}
	if (glst_parse_ipaddr(cldom, caddr, sizeof(caddr) - 1, &cport) < 0 ||
	    glst_parse_ipaddr(claddr, saddr, sizeof(saddr) - 1, &sport) < 0) {
		if ((svdom = strtok(NULL, tokstr)) == NULL ||
		    (svaddr = strtok(NULL, tokstr)) == NULL) {
			free(ldup);
			return -1;
		}
		if (glst_parse_ipaddr(claddr, caddr, sizeof(caddr) - 1, &cport) < 0 ||
		    glst_parse_ipaddr(svaddr, saddr, sizeof(saddr) - 1, &sport) < 0) {
			free(ldup);
			return -1;
		}
	}
	if (glst_assign_raddr(cfg, ctx, caddr) < 0) {
		free(ldup);
		return -1;
	}
	free(ctx->laddr);
	ctx->laddr = strdup(saddr);

	free(ldup);

	return 0;
}

static int glst_complete_ctx(char const *hblk, glst_cfg_t *cfg, glst_ctx_t *ctx) {
	int asize;
	char const *pos, *next, *addr;

	for (pos = hblk, next = NULL; (pos = glst_nextline(pos, &next)) != NULL;) {
		if (pos == hblk)
			glst_parse_xmail_addrline(pos, next - pos, cfg, ctx);
		if (strncasecmp(pos, MAIL_FROM, CSTRSIZE(MAIL_FROM)) == 0) {
			if (glst_get_address(pos, next - pos, &addr, &asize) == 0) {
				free(ctx->sender);
				ctx->sender = glst_deVERP_address(addr, asize);
			}
		} else if (strncasecmp(pos, RCPT_TO, CSTRSIZE(RCPT_TO)) == 0) {
			if (glst_get_address(pos, next - pos, &addr, &asize) == 0)
				glst_add_rcpt(ctx, addr, asize);
		} else if (strncasecmp(pos, MAIL_DATA_TAG, CSTRSIZE(MAIL_DATA_TAG)) == 0)
			break;
	}

	return 0;
}

static char *glst_ctx_first_key(glst_ctx_t *ctx) {
	int i, len;
	char *ckey, *rcpt;

	if ((ctx->crcpt = ctx->rcpts) == NULL)
		return NULL;
	rcpt = ctx->crcpt->ptr;
	len = strlen(ctx->mraddr) + strlen(ctx->sender) + strlen(rcpt) + 8;
	if ((ckey = (char *) malloc(len)) == NULL)
		return NULL;
	sprintf(ckey, "%s,%s,%s", ctx->mraddr, ctx->sender, rcpt);
	for (i = 0; ckey[i]; i++)
		if (isupper(ckey[i]))
			ckey[i] = tolower(ckey[i]);

	return ckey;
}

static char *glst_ctx_next_key(glst_ctx_t *ctx) {
	int i, len;
	char *ckey, *rcpt;

	if (ctx->crcpt == NULL || (ctx->crcpt = ctx->crcpt->next) == NULL)
		return NULL;
	rcpt = ctx->crcpt->ptr;
	len = strlen(ctx->mraddr) + strlen(ctx->sender) + strlen(rcpt) + 8;
	if ((ckey = (char *) malloc(len)) == NULL)
		return NULL;
	sprintf(ckey, "%s,%s,%s", ctx->mraddr, ctx->sender, rcpt);
	for (i = 0; ckey[i]; i++)
		if (isupper(ckey[i]))
			ckey[i] = tolower(ckey[i]);

	return ckey;
}

static int glst_is_complete_ctx(glst_ctx_t const *ctx) {

	return ctx->mraddr != NULL && ctx->raddr != NULL &&
		ctx->sender != NULL && ctx->rcpts != NULL;
}

static int glst_add_rcpt(glst_ctx_t *ctx, char const *rcpt, int size) {
	glst_ptrchain_t *nrcpt;

	if ((nrcpt = (glst_ptrchain_t *) malloc(sizeof(glst_ptrchain_t))) == NULL)
		return -1;
	nrcpt->ptr = glst_deVERP_address(rcpt, size);
	nrcpt->next = ctx->rcpts;
	ctx->rcpts = nrcpt;

	return 0;
}

static int glst_add_exnet(glst_cfg_t *cfg, char const *xnet, int size) {
	char *dxnet, *net, *mask;
	glst_subnet_t *cnet;
	glst_sockaddr_t anet, amask;

	if ((dxnet = glst_memdup(xnet, size, 0)) == NULL)
		return -1;
	if ((net = strtok(dxnet, ",")) == NULL ||
	    (mask = strtok(NULL, ",")) == NULL ||
	    glst_gethostbyname(net, GLST_INET46, 1, &anet) < 0 ||
	    glst_gethostbyname(mask, anet.a.sa_family, 1, &amask) < 0) {
		free(dxnet);
		return -1;
	}
	free(dxnet);

	if ((cnet = (glst_subnet_t *) malloc(sizeof(glst_subnet_t))) == NULL)
		return -1;
	cnet->net = anet;
	cnet->mask = amask;

	cnet->next = cfg->xnets;
	cfg->xnets = cnet;

	return 0;
}

static int glst_on_xnet(glst_cfg_t *cfg, glst_ctx_t *ctx) {
	glst_subnet_t *cnet;
	glst_sockaddr_t addr;

	if (!ctx->raddr ||
	    glst_gethostbyname(ctx->raddr, GLST_INET46, 1, &addr) < 0)
		return 0;
	for (cnet = cfg->xnets; cnet; cnet = cnet->next)
		if (glst_addr_match(&addr, &cnet->net, &cnet->mask))
			return 1;

	return 0;
}

static int glst_init_ctx(glst_ctx_t *ctx) {

	ctx->op = GLST_OP_GRAYLIST;
	ctx->raddr = ctx->mraddr = ctx->laddr = NULL;
	ctx->sender = strdup("null");
	ctx->rcpts = NULL;
	ctx->crcpt = NULL;
	ctx->mfile = NULL;
	ctx->rejfile = NULL;

	return 0;
}

static char *glst_load_hdrs(FILE *fin) {
	int i, nbh, hsize = 4096;
	char *hblk, *rhblk;
	char buf[1024];

	if ((hblk = (char *) malloc(hsize)) == NULL)
		return NULL;
	for (nbh = 0; fgets(buf, sizeof(buf) - 1, fin) != NULL;) {
		if ((i = strlen(buf)) > 0) {
			for (; i > 0 && strchr("\r\n", buf[i - 1]) != NULL; i--);
			buf[i] = 0;
		}
		if (!i)
			break;
		if (nbh + i + 1 >= hsize) {
			hsize = 2 * hsize + 1;
			if ((rhblk = (char *) realloc(hblk, hsize)) == NULL) {
				free(hblk);
				return NULL;
			}
			hblk = rhblk;
		}
		memcpy(hblk + nbh, buf, i);
		nbh += i;
		strcpy(hblk + nbh, "\n");
		nbh += 1;
	}
	hblk[nbh] = 0;

	return hblk;
}

static int glst_add_masknet(glst_cfg_t *cfg, char const *mnet, int size) {
	char *dmnet, *net, *mask, *nmask;
	glst_masknet_t *cnet;
	glst_sockaddr_t anet, amask, anmask;

	if ((dmnet = glst_memdup(mnet, size, 0)) == NULL)
		return -1;
	if ((net = strtok(dmnet, ",")) == NULL ||
	    (mask = strtok(NULL, ",")) == NULL ||
	    (nmask = strtok(NULL, ",")) == NULL ||
	    glst_gethostbyname(net, GLST_INET46, 1, &anet) < 0 ||
	    glst_gethostbyname(mask, anet.a.sa_family, 1, &amask) < 0 ||
	    glst_gethostbyname(nmask, anet.a.sa_family, 1, &anmask) < 0) {
		free(dmnet);
		return -1;
	}
	free(dmnet);

	if ((cnet = (glst_masknet_t *) malloc(sizeof(glst_masknet_t))) == NULL)
		return -1;
	cnet->net = anet;
	cnet->mask = amask;
	cnet->nmask = anmask;

	cnet->next = cfg->mnets;
	cfg->mnets = cnet;

	return 0;
}

static char *glst_get_masked_raddr(glst_cfg_t *cfg, char const *raddr) {
	int i, asize, nsize;
	unsigned char *adata, *ndata;
	glst_masknet_t *cnet;
	glst_sockaddr_t addr;
	char ipaddr[256];

	if (glst_gethostbyname(raddr, GLST_INET46, 1, &addr) < 0)
		return strdup(raddr);
	for (cnet = cfg->mnets; cnet; cnet = cnet->next) {
		if (glst_addr_match(&addr, &cnet->net, &cnet->mask)) {
			if (glst_gethostbyname(raddr, cnet->nmask.a.sa_family,
					       1, &addr) < 0 ||
			    (adata = (unsigned char *)
			     glst_addr_data(&addr, &asize)) == NULL ||
			    (ndata = (unsigned char *)
			     glst_addr_data(&cnet->nmask, &nsize)) == NULL)
				return strdup(raddr);
			for (i = 0; i < asize; i++)
				adata[i] &= ndata[i];
			glst_inet_ntoa(&addr, ipaddr, sizeof(ipaddr));

			return strdup(ipaddr);
		}
	}

	return strdup(raddr);
}

static int glst_assign_raddr(glst_cfg_t *cfg, glst_ctx_t *ctx, char const *raddr) {

	free(ctx->raddr);
	free(ctx->mraddr);
	ctx->raddr = strdup(raddr);
	ctx->mraddr = glst_get_masked_raddr(cfg, raddr);

	return 0;
}

static int glst_set_raddr(glst_cfg_t *cfg, glst_ctx_t *ctx, char const *raddr) {
	int aport;
	char ipaddr[256];

	if (glst_parse_ipaddr(raddr, ipaddr, sizeof(ipaddr) - 1, &aport) < 0 ||
	    glst_assign_raddr(cfg, ctx, ipaddr) < 0)
		return -1;

	return 0;
}

static int glst_init(int ac, char **av) {
#if defined(WIN32)
	WSADATA wsd;

	memset(&wsd, 0, sizeof(wsd));
	if (WSAStartup(MAKEWORD(2, 0), &wsd))
		return -1;
#endif

	return 0;
}

static int glst_init_config(char const *prg, glst_cfg_t *cfg) {
	char const *tmp, *rootd = ".";
	char root[512];

	cfg->generror = GLST_GENERROR;
	cfg->rejerror = GLST_REJERROR;
	cfg->timeo = GLST_DEF_TIMEO;
	cfg->exptimeo = GLST_DEF_EXPTIMEO;
	cfg->lametimeo = GLST_DEF_LAMETIMEO;
	cfg->rejmsg = strdup(GLST_DEF_REJMSG);
	cfg->xnets = NULL;
	cfg->mnets = NULL;

	if ((tmp = strrchr(prg, SLASH_CHAR)) != NULL &&
	    (tmp - prg) + 1 < sizeof(root)) {
		strncpy(root, prg, tmp - prg);
		root[tmp - prg] = 0;
		rootd = root;
	}
	cfg->root = strdup(rootd);

	return 0;
}

static int glst_load_config(char const *prg, char const *fcfg, glst_cfg_t *cfg) {
	char const *cfgfile = fcfg;
	FILE *cfil;
	char *tmp, *var, *val;
	char fcfgbuf[512], buf[512];

	if (cfgfile == NULL) {
		snprintf(fcfgbuf, sizeof(fcfgbuf) - 1, "%s%s%s",
			 cfg->root, SLASH_STR, GLST_CONFIG_FILE);
		cfgfile = fcfgbuf;
	}
	if ((cfil = fopen(cfgfile, "r")) == NULL) {
		if (fcfg)
			perror(fcfg);
		return -1;
	}

	while (fgets(buf, sizeof(buf) - 1, cfil) != NULL) {
		buf[strlen(buf) - 1] = 0;
		for (var = buf; *var && strchr(" \t", *var); var++);
		if (!*var)
			continue;
		for (tmp = var; *tmp && (isalpha(*tmp) || isdigit(*tmp)); tmp++);
		if (!*tmp || tmp == var || strchr(" \t=", *tmp) == NULL)
			continue;
		for (val = tmp; *val && strchr(" \t", *val) != NULL; val++);
		if (*val++ != '=')
			continue;
		*tmp++ = 0;
		for (; *val && strchr(" \t", *val) != NULL; val++);
		for (tmp = val + strlen(val); tmp > val &&
			     strchr(" \t\r\n", tmp[-1]) != NULL; tmp--);
		*tmp = 0;

		if (!strcmp(var, "root")) {
			free(cfg->root);
			cfg->root = strdup(val);
		} else if (!strcmp(var, "rejmsg")) {
			free(cfg->rejmsg);
			cfg->rejmsg = strdup(val);
		} else if (!strcmp(var, "generr")) {
			cfg->generror = atoi(val);
		} else if (!strcmp(var, "rejerr")) {
			cfg->rejerror = atoi(val);
		} else if (!strcmp(var, "timeo")) {
			cfg->timeo = atoi(val);
		} else if (!strcmp(var, "exptimeo")) {
			cfg->exptimeo = atoi(val);
		} else if (!strcmp(var, "lametimeo")) {
			cfg->lametimeo = atoi(val);
		} else if (!strcmp(var, "xnet")) {
			glst_add_exnet(cfg, val, -1);
		} else if (!strcmp(var, "mnet")) {
			glst_add_masknet(cfg, val, -1);
		}
	}
	fclose(cfil);

	return 0;
}

static int glst_get_dbvar(char const *data, int dsize, char const *name,
			  char *vbuf, int vsize) {
	int j, nlen = strlen(name);
	char const *pos, *var;

	for (pos = data; pos < (data + dsize);) {
		if ((var = strstr(pos, name)) == NULL)
			break;
		if (var[nlen] != '=') {
			pos = var + 1;
			continue;
		}
		for (j = 0, var += nlen + 1;
		     j < vsize && var < (data + dsize); j++, var++) {
			if (*var == '\\' && (var + 1) < (data + dsize)) {
				var++;
				vbuf[j] = *var;
			} else if (*var == ',') {
				break;
			} else {
				vbuf[j] = *var;
			}
		}
		if (j == vsize)
			return -1;
		vbuf[j] = 0;
		return 0;
	}

	return -1;
}

static int glst_init_glrecord(glst_glrecord_t *rec) {
	time_t tcur = time(NULL);

	rec->rtime = tcur;
	rec->mtime = tcur;
	rec->mcount = 0;
	rec->acount = 1;

	return 0;
}

static void glst_free_glrecord(glst_glrecord_t *rec) {


}

static int glst_parse_glrecord(datum ddata, glst_glrecord_t *rec) {
	unsigned long rtime;
	char var[1024];

	if (glst_get_dbvar(ddata.dptr, ddata.dsize, "rtime", var,
			   sizeof(var)) < 0 ||
	    sscanf(var, "%lx", &rtime) != 1) {
		fprintf(stderr, "bad record format: rec='%s'\n",
			ddata.dptr);
		return -1;
	}
	rec->rtime = (time_t) rtime;

	if (glst_get_dbvar(ddata.dptr, ddata.dsize, "mtime", var,
			   sizeof(var)) < 0 ||
	    sscanf(var, "%lx", &rtime) != 1) {
		fprintf(stderr, "bad record format: rec='%s'\n",
			ddata.dptr);
		return -1;
	}
	rec->mtime = (time_t) rtime;

	if (glst_get_dbvar(ddata.dptr, ddata.dsize, "mcnt", var,
			   sizeof(var)) < 0 ||
	    sscanf(var, "%d", &rec->mcount) != 1) {
		fprintf(stderr, "bad record format: rec='%s'\n",
			ddata.dptr);
		return -1;
	}

	if (glst_get_dbvar(ddata.dptr, ddata.dsize, "acnt", var,
			   sizeof(var)) < 0 ||
	    sscanf(var, "%d", &rec->acount) != 1) {
		fprintf(stderr, "bad record format: rec='%s'\n",
			ddata.dptr);
		return -1;
	}

	return 0;
}

static int glst_insert_glrecord(GDBM_FILE gdbf, datum dkey, glst_glrecord_t const *rec) {
	char *dbrec;
	datum ddata;

	if ((dbrec = glst_sprintf("rtime=%lx,mtime=%lx,mcnt=%d,acnt=%d",
				  (unsigned long) rec->rtime, (unsigned long) rec->mtime,
				  rec->mcount, rec->acount)) == NULL)
		return -1;
	ddata.dptr = dbrec;
	ddata.dsize = strlen(dbrec) + 1;
	if (gdbm_store(gdbf, dkey, ddata, GDBM_REPLACE) != 0) {
		fprintf(stderr, "unable to store to DB file\n");
		free(dbrec);
		return -1;
	}
	free(dbrec);

	return 0;
}

static int glst_init_lmrecord(glst_lmrecord_t *rec) {
	time_t tcur = time(NULL);

	rec->rtime = tcur;
	rec->mtime = tcur;
	rec->count = 1;

	return 0;
}

static void glst_free_lmrecord(glst_lmrecord_t *rec) {


}

static int glst_parse_lmrecord(datum ddata, glst_lmrecord_t *rec) {
	unsigned long rtime;
	char var[1024];

	if (glst_get_dbvar(ddata.dptr, ddata.dsize, "rtime", var,
			   sizeof(var)) < 0 ||
	    sscanf(var, "%lx", &rtime) != 1) {
		fprintf(stderr, "bad record format: rec='%s'\n",
			ddata.dptr);
		return -1;
	}
	rec->rtime = (time_t) rtime;

	if (glst_get_dbvar(ddata.dptr, ddata.dsize, "mtime", var,
			   sizeof(var)) < 0 ||
	    sscanf(var, "%lx", &rtime) != 1) {
		fprintf(stderr, "bad record format: rec='%s'\n",
			ddata.dptr);
		return -1;
	}
	rec->mtime = (time_t) rtime;

	if (glst_get_dbvar(ddata.dptr, ddata.dsize, "cnt", var,
			   sizeof(var)) < 0 ||
	    sscanf(var, "%d", &rec->count) != 1) {
		fprintf(stderr, "bad record format: rec='%s'\n",
			ddata.dptr);
		return -1;
	}

	return 0;
}

static int glst_insert_lmrecord(GDBM_FILE gdbf, datum dkey, glst_lmrecord_t const *rec) {
	char *dbrec;
	datum ddata;

	if ((dbrec = glst_sprintf("rtime=%lx,mtime=%lx,cnt=%d",
				  (unsigned long) rec->rtime, (unsigned long) rec->mtime,
				  rec->count)) == NULL)
		return -1;
	ddata.dptr = dbrec;
	ddata.dsize = strlen(dbrec) + 1;
	if (gdbm_store(gdbf, dkey, ddata, GDBM_REPLACE) != 0) {
		fprintf(stderr, "unable to store to DB file\n");
		free(dbrec);
		return -1;
	}
	free(dbrec);

	return 0;
}

static int glst_handle_recipient(GDBM_FILE gdbf, char *ckey, glst_ctx_t *ctx,
				 glst_cfg_t *cfg) {
	int res = -1;
	time_t tcur = time(NULL);
	datum dkey, ddata;
	glst_glrecord_t rec;

	dkey.dptr = ckey;
	dkey.dsize = strlen(ckey) + 1;
	ddata = gdbm_fetch(gdbf, dkey);
	if (ddata.dptr != NULL) {
		if (glst_parse_glrecord(ddata, &rec) < 0) {
			free(ddata.dptr);
			return -1;
		}
		if ((tcur - rec.rtime) > (time_t) cfg->timeo) {
			rec.mtime = tcur;
			rec.mcount++;
			if (glst_insert_glrecord(gdbf, dkey, &rec) == 0)
				res = 0;
		} else {
			rec.mtime = tcur;
			rec.acount++;
			glst_insert_glrecord(gdbf, dkey, &rec);
			res = 1;
		}
		glst_free_glrecord(&rec);
		free(ddata.dptr);
	} else {
		if (glst_init_glrecord(&rec) < 0)
			return -1;
		if (glst_insert_glrecord(gdbf, dkey, &rec) == 0)
			res = 1;
		glst_free_glrecord(&rec);
	}

	return res;
}

static int glst_handle_request(glst_ctx_t *ctx, glst_cfg_t *cfg) {
	int res, rpos = 0, rneg = 0;
	char *ckey;
	GDBM_FILE gdbf;
	char dbfile[512];

	snprintf(dbfile, sizeof(dbfile) - 1, "%s%s%s", cfg->root, SLASH_STR, GLST_DB_FILE);
	if ((gdbf = gdbm_open(dbfile, 0, GDBM_WRCREAT, 00664, NULL)) == NULL) {
		fprintf(stderr, "unable to open DB file: file='%s'\n", dbfile);
		return -1;
	}
	for (ckey = glst_ctx_first_key(ctx); ckey; ckey = glst_ctx_next_key(ctx)) {
		if ((res = glst_handle_recipient(gdbf, ckey, ctx, cfg)) > 0)
			rpos++;
		else if (res < 0)
			rneg++;
		free(ckey);
	}
	gdbm_close(gdbf);

	return rneg ? -1: (rpos ? 1: 0);
}

static int glst_add_keychain(glst_keychain_t **keys, char const *key, char const *data) {
	glst_keychain_t *kv;

	if ((kv = (glst_keychain_t *) malloc(sizeof(glst_keychain_t))) == NULL)
		return -1;
	kv->key = strdup(key);
	kv->data = strdup(data);
	kv->next = *keys;
	*keys = kv;

	return 0;
}

static void glst_free_keychain(glst_keychain_t *kv) {

	free(kv->data);
	free(kv->key);
	free(kv);
}

static void glst_free_keychain_list(glst_keychain_t *keys) {
	glst_keychain_t *kv;

	for (; (kv = keys) != NULL;) {
		keys = keys->next;
		glst_free_keychain(kv);
	}
}

static int glst_update_lamedb(glst_cfg_t *cfg, glst_keychain_t *keys) {
	GDBM_FILE gdbf;
	time_t tcur = time(NULL);
	char *ckey, *ipa, *sender, *rcpt;
	char const *tokstr = ",";
	glst_keychain_t *kv;
	glst_lmrecord_t lrec;
	datum dkey, ddata;
	char dbfile[512];

	snprintf(dbfile, sizeof(dbfile) - 1, "%s%s%s", cfg->root, SLASH_STR, GLST_LAMEDB_FILE);
	if ((gdbf = gdbm_open(dbfile, 0, GDBM_WRCREAT, 00664, NULL)) == NULL) {
		fprintf(stderr, "unable to open DB file: file='%s'\n", dbfile);
		return -1;
	}
	for (kv = keys; kv != NULL; kv = kv->next) {
		if ((ckey = strdup(kv->key)) == NULL) {
			gdbm_close(gdbf);
			return -1;
		}
		if ((ipa = strtok(ckey, tokstr)) == NULL ||
		    (sender = strtok(NULL, tokstr)) == NULL ||
		    (rcpt = strtok(NULL, tokstr)) == NULL) {
			free(ckey);
			gdbm_close(gdbf);
			return -1;
		}
		dkey.dptr = ipa;
		dkey.dsize = strlen(ipa) + 1;
		ddata = gdbm_fetch(gdbf, dkey);
		if (ddata.dptr != NULL) {
			if (glst_parse_lmrecord(ddata, &lrec) == 0) {
				lrec.mtime = tcur;
				lrec.count++;
			} else
				glst_init_lmrecord(&lrec);
			free(ddata.dptr);
		} else
			glst_init_lmrecord(&lrec);

		glst_insert_lmrecord(gdbf, dkey, &lrec);
		glst_free_lmrecord(&lrec);

		free(ckey);
	}
	gdbm_close(gdbf);

	return 0;
}

static int glst_handle_cleanup(glst_cfg_t *cfg) {
	int res, recdel;
	GDBM_FILE gdbf;
	time_t tcur = time(NULL);
	glst_keychain_t *lmkeys = NULL, *exkeys = NULL, *kv;
	datum dkey, ddata;
	glst_glrecord_t rec;
	char dbfile[512];

	snprintf(dbfile, sizeof(dbfile) - 1, "%s%s%s", cfg->root, SLASH_STR, GLST_DB_FILE);
	if ((gdbf = gdbm_open(dbfile, 0, GDBM_WRCREAT, 00664, NULL)) == NULL) {
		fprintf(stderr, "unable to open DB file: file='%s'\n", dbfile);
		return -1;
	}
	ddata = gdbm_firstkey(gdbf);
	while (ddata.dptr != NULL) {
		dkey = ddata;
		ddata = gdbm_fetch(gdbf, dkey);
		if (ddata.dptr != NULL) {
			if (glst_parse_glrecord(ddata, &rec) < 0) {
				glst_add_keychain(&lmkeys, dkey.dptr, ddata.dptr);
			} else {
				if ((tcur - rec.mtime) > (time_t) cfg->exptimeo)
					glst_add_keychain(&exkeys, dkey.dptr, ddata.dptr);
				if (rec.mcount == 0 && (tcur - rec.mtime) > (time_t) cfg->lametimeo)
					glst_add_keychain(&lmkeys, dkey.dptr, ddata.dptr);
				glst_free_glrecord(&rec);
			}
			free(ddata.dptr);
		}
		ddata = gdbm_nextkey(gdbf, dkey);
		free(dkey.dptr);
	}
	for (recdel = 0, kv = lmkeys; kv != NULL; kv = kv->next) {
		dkey.dptr = kv->key;
		dkey.dsize = strlen(kv->key) + 1;
		if (gdbm_delete(gdbf, dkey) != 0) {
			fprintf(stderr, "unable to delete record: key='%s'\n",
				kv->key);
		} else
			recdel++;
	}
	for (kv = exkeys; kv != NULL; kv = kv->next) {
		dkey.dptr = kv->key;
		dkey.dsize = strlen(kv->key) + 1;
		if (gdbm_delete(gdbf, dkey) != 0) {
			fprintf(stderr, "unable to delete record: key='%s'\n",
				kv->key);
		} else
			recdel++;
	}
	if (recdel)
		gdbm_reorganize(gdbf);
	gdbm_close(gdbf);
	res = glst_update_lamedb(cfg, lmkeys);
	glst_free_keychain_list(lmkeys);
	glst_free_keychain_list(exkeys);

	return 0;
}

static void glst_usage(char const *prg) {

	fprintf(stderr,
		"use: %s [--mfile MFIL] [--raddr RADDR] [--sender SNDR] [--rcpt RCPT]\n"
		"\t[--generr ERR] [--rejerr ERR] [--timeo NSEC] [--cleanup] [--xnet NET]\n"
		"\t[--mnet MNET] [--exptimeo NSEC] [--lametimeo NSEC] [--cfg CFIL]\n"
		"\t[--rejfile RFIL] [--help]\n", prg);
}

int main(int ac, char **av) {
	int i, res;
	glst_ctx_t ctx;
	glst_cfg_t cfg;

	if (glst_init(ac, av) < 0)
		return GLST_GENERROR;
	glst_init_config(av[0], &cfg);
	glst_init_ctx(&ctx);
	glst_load_config(av[0], NULL, &cfg);
	for (i = 1; i < ac; i++) {
		if (!strcmp(av[i], "--cfg")) {
			if (++i < ac &&
			    glst_load_config(av[0], av[i], &cfg) < 0) {

				return cfg.generror;
			}
		} else if (!strcmp(av[i], "--raddr")) {
			if (++i < ac)
				glst_set_raddr(&cfg, &ctx, av[i]);
		} else if (!strcmp(av[i], "--sender")) {
			if (++i < ac) {
				free(ctx.sender);
				ctx.sender = strdup(av[i]);
			}
		} else if (!strcmp(av[i], "--rcpt")) {
			if (++i < ac)
				glst_add_rcpt(&ctx, av[i], -1);
		} else if (!strcmp(av[i], "--mfile")) {
			if (++i < ac) {
				free(ctx.mfile);
				ctx.mfile = strdup(av[i]);
			}
		} else if (!strcmp(av[i], "--rejfile")) {
			if (++i < ac) {
				free(ctx.rejfile);
				ctx.rejfile = strdup(av[i]);
			}
		} else if (!strcmp(av[i], "--generr")) {
			if (++i < ac)
				cfg.generror = atoi(av[i]);
		} else if (!strcmp(av[i], "--rejerr")) {
			if (++i < ac)
				cfg.rejerror = atoi(av[i]);
		} else if (!strcmp(av[i], "--timeo")) {
			if (++i < ac)
				cfg.timeo = atoi(av[i]);
		} else if (!strcmp(av[i], "--exptimeo")) {
			if (++i < ac)
				cfg.exptimeo = atoi(av[i]);
		} else if (!strcmp(av[i], "--lametimeo")) {
			if (++i < ac)
				cfg.lametimeo = atoi(av[i]);
		} else if (!strcmp(av[i], "--xnet")) {
			if (++i < ac)
				glst_add_exnet(&cfg, av[i], -1);
		} else if (!strcmp(av[i], "--mnet")) {
			if (++i < ac)
				glst_add_masknet(&cfg, av[i], -1);
		} else if (!strcmp(av[i], "--cleanup")) {
			ctx.op = GLST_OP_CLEANUP;
		} else {
			glst_usage(av[0]);
			return cfg.generror;
		}
	}
	if (ctx.op == GLST_OP_CLEANUP) {
		res = glst_handle_cleanup(&cfg);

		return res == 0 ? 0: cfg.generror;
	}

	if (!glst_is_complete_ctx(&ctx)) {
		char *hblk;
		FILE *fin = stdin;

		if (ctx.mfile != NULL && (fin = fopen(ctx.mfile, "r")) == NULL) {
			perror(ctx.mfile);
			return cfg.generror;
		}
		hblk = glst_load_hdrs(fin);
		if (fin != stdin)
			fclose(fin);
		if (!hblk)
			return cfg.generror;
		res = glst_complete_ctx(hblk, &cfg, &ctx);
		free(hblk);
		if (res < 0 || !glst_is_complete_ctx(&ctx))
			return cfg.generror;
	}

	if (glst_on_xnet(&cfg, &ctx))
		return 0;

	if ((res = glst_handle_request(&ctx, &cfg)) > 0) {
		FILE *fout = stdout;
		char *rejfile = ctx.rejfile;
		char rbuf[512];

		if (rejfile == NULL && ctx.mfile != NULL) {
			snprintf(rbuf, sizeof(rbuf) - 1, "%s.rej", ctx.mfile);
			rejfile = rbuf;
		}
		if (rejfile != NULL && (fout = fopen(rejfile, "w")) == NULL) {
			perror(rejfile);
			return cfg.generror;
		}
		fprintf(fout, "%s\n", cfg.rejmsg);
		if (fout != stdout)
			fclose(fout);

		return cfg.rejerror;
	}

	return res < 0 ? cfg.generror: 0;
}

